"""
Common Utilities
================

Defines the common utilities objects that don't fall in any specific category.

References
----------
-   :cite:`Kienzle2011a` : Kienzle, P., Patel, N., & Krycka, J. (2011).
    refl1d.numpyerrors - Refl1D v0.6.19 documentation. Retrieved January 30,
    2015, from
    http://www.reflectometry.org/danse/docs/refl1d/_modules/refl1d/\
numpyerrors.html
"""

from __future__ import annotations

import inspect
import multiprocessing
import multiprocessing.pool
import functools
import numpy as np
import re
import types
import warnings
from contextlib import contextmanager
from copy import copy
from pprint import pformat

from colour.constants import INTEGER_THRESHOLD
from colour.hints import (
    Any,
    Boolean,
    Callable,
    Dict,
    Generator,
    Integer,
    Iterable,
    Literal,
    Mapping,
    Optional,
    RegexFlag,
    Sequence,
    TypeVar,
    Union,
)
from colour.utilities import CaseInsensitiveMapping, Lookup

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "CacheRegistry",
    "CACHE_REGISTRY",
    "handle_numpy_errors",
    "ignore_numpy_errors",
    "raise_numpy_errors",
    "print_numpy_errors",
    "warn_numpy_errors",
    "ignore_python_warnings",
    "attest",
    "batch",
    "disable_multiprocessing",
    "multiprocessing_pool",
    "is_matplotlib_installed",
    "is_networkx_installed",
    "is_opencolorio_installed",
    "is_openimageio_installed",
    "is_pandas_installed",
    "is_sklearn_installed",
    "is_tqdm_installed",
    "is_trimesh_installed",
    "required",
    "is_iterable",
    "is_string",
    "is_numeric",
    "is_integer",
    "is_sibling",
    "filter_kwargs",
    "filter_mapping",
    "first_item",
    "copy_definition",
    "validate_method",
    "optional",
]


class CacheRegistry:
    """
    A registry for mapping-based caches.

    Attributes
    ----------
    -   :attr:`~colour.utilities.CacheRegistry.registry`

    Methods
    -------
    -   :meth:`~colour.SpectralShape.__init__`
    -   :meth:`~colour.SpectralShape.__str__`
    -   :meth:`~colour.SpectralShape.register_cache`
    -   :meth:`~colour.SpectralShape.unregister_cache`
    -   :meth:`~colour.SpectralShape.clear_cache`
    -   :meth:`~colour.SpectralShape.clear_all_caches`

    Examples
    --------
    >>> cache_registry = CacheRegistry()
    >>> cache_a = cache_registry.register_cache('Cache A')
    >>> cache_a['Foo'] = 'Bar'
    >>> cache_b = cache_registry.register_cache('Cache B')
    >>> cache_b['John'] = 'Doe'
    >>> cache_b['Luke'] = 'Skywalker'
    >>> print(cache_registry)
    {'Cache A': '1 item(s)', 'Cache B': '2 item(s)'}
    >>> cache_registry.clear_cache('Cache A')
    >>> print(cache_registry)
    {'Cache A': '0 item(s)', 'Cache B': '2 item(s)'}
    >>> cache_registry.unregister_cache('Cache B')
    >>> print(cache_registry)
    {'Cache A': '0 item(s)'}
    >>> print(cache_b)
    {}
    """

    def __init__(self):
        self._registry = {}

    @property
    def registry(self) -> Dict:
        """
        Getter property for the cache registry.

        Returns
        -------
        :class:`dict`
            Cache registry.
        """

        return self._registry

    def __str__(self) -> str:
        """
        Return a formatted string representation of the cache registry.

        Returns
        -------
        :class:`str`
            Formatted string representation.
        """

        return pformat(
            {
                name: f"{len(self._registry[name])} item(s)"
                for name in sorted(self._registry)
            }
        )

    def register_cache(self, name: str) -> Dict:
        """
        Register a new cache with given name in the registry.

        Parameters
        ----------
        name
            Cache name for the registry.

        Returns
        -------
        :class:`dict`
            Registered cache.

        Examples
        --------
        >>> cache_registry = CacheRegistry()
        >>> cache_a = cache_registry.register_cache('Cache A')
        >>> cache_a['Foo'] = 'Bar'
        >>> cache_b = cache_registry.register_cache('Cache B')
        >>> cache_b['John'] = 'Doe'
        >>> cache_b['Luke'] = 'Skywalker'
        >>> print(cache_registry)
        {'Cache A': '1 item(s)', 'Cache B': '2 item(s)'}
        """

        self._registry[name] = {}

        return self._registry[name]

    def unregister_cache(self, name: str):
        """
        Unregister cache with given name in the registry.

        Parameters
        ----------
        name
            Cache name in the registry.

        Notes
        -----
        -   The cache is cleared before being unregistered.

        Examples
        --------
        >>> cache_registry = CacheRegistry()
        >>> cache_a = cache_registry.register_cache('Cache A')
        >>> cache_a['Foo'] = 'Bar'
        >>> cache_b = cache_registry.register_cache('Cache B')
        >>> cache_b['John'] = 'Doe'
        >>> cache_b['Luke'] = 'Skywalker'
        >>> print(cache_registry)
        {'Cache A': '1 item(s)', 'Cache B': '2 item(s)'}
        >>> cache_registry.unregister_cache('Cache B')
        >>> print(cache_registry)
        {'Cache A': '1 item(s)'}
        >>> print(cache_b)
        {}
        """

        self.clear_cache(name)

        del self._registry[name]

    def clear_cache(self, name: str):
        """
        Clear the cache with given name.

        Parameters
        ----------
        name
            Cache name in the registry.

        Examples
        --------
        >>> cache_registry = CacheRegistry()
        >>> cache_a = cache_registry.register_cache('Cache A')
        >>> cache_a['Foo'] = 'Bar'
        >>> print(cache_registry)
        {'Cache A': '1 item(s)'}
        >>> cache_registry.clear_cache('Cache A')
        >>> print(cache_registry)
        {'Cache A': '0 item(s)'}
        """

        self._registry[name].clear()

    def clear_all_caches(self):
        """
        Clear all the caches in the registry.

        Examples
        --------
        >>> cache_registry = CacheRegistry()
        >>> cache_a = cache_registry.register_cache('Cache A')
        >>> cache_a['Foo'] = 'Bar'
        >>> cache_b = cache_registry.register_cache('Cache B')
        >>> cache_b['John'] = 'Doe'
        >>> cache_b['Luke'] = 'Skywalker'
        >>> print(cache_registry)
        {'Cache A': '1 item(s)', 'Cache B': '2 item(s)'}
        >>> cache_registry.clear_all_caches()
        >>> print(cache_registry)
        {'Cache A': '0 item(s)', 'Cache B': '0 item(s)'}
        """

        for key in self._registry:
            self.clear_cache(key)


CACHE_REGISTRY: CacheRegistry = CacheRegistry()
"""
*Colour* cache registry referencing all the caches used for repetitive or long
processes.

CACHE_REGISTRY
"""


def handle_numpy_errors(**kwargs: Any) -> Callable:
    """
    Decorate a function to handle *Numpy* errors.

    Other Parameters
    ----------------
    kwargs
        Keywords arguments.

    Returns
    -------
    Callable

    References
    ----------
    :cite:`Kienzle2011a`

    Examples
    --------
    >>> import numpy
    >>> @handle_numpy_errors(all='ignore')
    ... def f():
    ...     1 / numpy.zeros(3)
    >>> f()
    """

    context = np.errstate(**kwargs)

    def wrapper(function: Callable) -> Callable:
        """Wrap given function wrapper."""

        @functools.wraps(function)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            """Wrap given function."""

            with context:
                return function(*args, **kwargs)

        return wrapped

    return wrapper


ignore_numpy_errors = handle_numpy_errors(all="ignore")
raise_numpy_errors = handle_numpy_errors(all="raise")
print_numpy_errors = handle_numpy_errors(all="print")
warn_numpy_errors = handle_numpy_errors(all="warn")


def ignore_python_warnings(function: Callable) -> Callable:
    """
    Decorate a function to ignore *Python* warnings.

    Parameters
    ----------
    function
        Function to decorate.

    Returns
    -------
    Callable

    Examples
    --------
    >>> @ignore_python_warnings
    ... def f():
    ...     warnings.warn('This is an ignored warning!')
    >>> f()
    """

    @functools.wraps(function)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        """Wrap given function."""

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            return function(*args, **kwargs)

    return wrapper


def attest(condition: Boolean, message: str = ""):
    """
    Provide the `assert` statement functionality without being disabled by
    optimised Python execution.

    Parameters
    ----------
    condition
        Condition to attest/assert.
    message
        Message to display when the assertion fails.
    """

    if not condition:
        raise AssertionError(message)


def batch(sequence: Sequence, k: Union[Integer, Literal[3]] = 3) -> Generator:
    """
    Return a batch generator from given sequence.

    Parameters
    ----------
    sequence
        Sequence to create batches from.
    k
        Batch size.

    Yields
    ------
    Generator
        Batch generator.

    Examples
    --------
    >>> batch(tuple(range(10)), 3)  # doctest: +ELLIPSIS
    <generator object batch at 0x...>
    """

    for i in range(0, len(sequence), k):
        yield sequence[i : i + k]


_MULTIPROCESSING_ENABLED: Boolean = True
"""*Colour* multiprocessing state."""


class disable_multiprocessing:
    """
    Define a context manager and decorator to temporarily disabling *Colour*
    multiprocessing state.
    """

    def __enter__(self) -> disable_multiprocessing:
        """
        Disable *Colour* multiprocessing state upon entering the context
        manager.
        """

        global _MULTIPROCESSING_ENABLED

        _MULTIPROCESSING_ENABLED = False

        return self

    def __exit__(self, *args: Any):
        """
        Enable *Colour* multiprocessing state upon exiting the context
        manager.
        """

        global _MULTIPROCESSING_ENABLED

        _MULTIPROCESSING_ENABLED = True

    def __call__(self, function: Callable) -> Callable:
        """Call the wrapped definition."""

        @functools.wraps(function)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            """Wrap given function."""

            with self:
                return function(*args, **kwargs)

        return wrapper


def _initializer(kwargs: Any):
    """
    Initialize a multiprocessing pool.

    It is used to ensure that processes on *Windows* inherit correctly from the
    current domain-range scale.

    Parameters
    ----------
    kwargs
        Initialisation arguments.
    """

    # NOTE: No coverage information is available as this code is executed in
    # sub-processes.

    import colour.utilities.array  # pragma: no cover

    colour.utilities.array._DOMAIN_RANGE_SCALE = kwargs.get(
        "scale", "reference"
    )  # pragma: no cover


@contextmanager
def multiprocessing_pool(*args: Any, **kwargs: Any) -> Generator:
    """
    Define a context manager providing a multiprocessing pool.

    Other Parameters
    ----------------
    args
        Arguments.
    kwargs
        Keywords arguments.

    Yields
    ------
    Generator
        Multiprocessing pool.

    Examples
    --------
    >>> from functools import partial
    >>> def _add(a, b):
    ...     return a + b
    >>> with multiprocessing_pool() as pool:
    ...     pool.map(partial(_add, b=2), range(10))
    ... # doctest: +SKIP
    [2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
    """

    from colour.utilities import get_domain_range_scale

    class _DummyPool:
        """
        A dummy multiprocessing pool that does not perform multiprocessing.

        Other Parameters
        ----------------
        args
            Arguments.
        kwargs
            Keywords arguments.
        """

        def __init__(self, *args: Any, **kwargs: Any):
            pass

        def map(self, func, iterable, chunksize=None):
            """Apply given function to each element of given iterable."""

            return [func(a) for a in iterable]

        def terminate(self):
            """Terminate the process."""

            pass

    kwargs["initializer"] = _initializer
    kwargs["initargs"] = ({"scale": get_domain_range_scale()},)

    pool_factory: Callable
    if _MULTIPROCESSING_ENABLED:
        pool_factory = multiprocessing.Pool
    else:
        pool_factory = _DummyPool

    pool = pool_factory(*args, **kwargs)

    try:
        yield pool
    finally:
        pool.terminate()


def is_matplotlib_installed(raise_exception: Boolean = False) -> Boolean:
    """
    Return whether *Matplotlib* is installed and available.

    Parameters
    ----------
    raise_exception
        Whether to raise an exception if *Matplotlib* is unavailable.

    Returns
    -------
    :class:`bool`
        Whether *Matplotlib* is installed.

    Raises
    ------
    :class:`ImportError`
        If *Matplotlib* is not installed.
    """

    try:  # pragma: no cover
        # pylint: disable=W0612
        import matplotlib  # noqa

        return True
    except ImportError as error:  # pragma: no cover
        if raise_exception:
            raise ImportError(
                '"Matplotlib" related API features are not available: '
                f'"{error}".\nSee the installation guide for more information: '
                "https://www.colour-science.org/installation-guide/"
            )

        return False


def is_networkx_installed(raise_exception: Boolean = False) -> Boolean:
    """
    Return whether *NetworkX* is installed and available.

    Parameters
    ----------
    raise_exception
        Whether to raise an exception if *NetworkX* is unavailable.

    Returns
    -------
    :class:`bool`
        Whether *NetworkX* is installed.

    Raises
    ------
    :class:`ImportError`
        If *NetworkX* is not installed.
    """

    try:  # pragma: no cover
        # pylint: disable=W0612
        import networkx  # noqa

        return True
    except ImportError as error:  # pragma: no cover
        if raise_exception:
            raise ImportError(
                '"NetworkX" related API features, e.g. the automatic colour '
                f'conversion graph, are not available: "{error}".\nPlease refer '
                "to the installation guide for more information: "
                "https://www.colour-science.org/installation-guide/"
            )

        return False


def is_opencolorio_installed(raise_exception: Boolean = False) -> Boolean:
    """
    Return whether *OpenColorIO* is installed and available.

    Parameters
    ----------
    raise_exception
        Whether to raise an exception if *OpenColorIO* is unavailable.

    Returns
    -------
    :class:`bool`
        Whether *OpenColorIO* is installed.

    Raises
    ------
    :class:`ImportError`
        If *OpenColorIO* is not installed.
    """

    try:  # pragma: no cover
        # pylint: disable=W0612
        import PyOpenColorIO  # noqa

        return True
    except ImportError as error:  # pragma: no cover
        if raise_exception:
            raise ImportError(
                '"OpenColorIO" related API features are not available: '
                f'"{error}".\nSee the installation guide for more information: '
                "https://www.colour-science.org/installation-guide/"
            )

        return False


def is_openimageio_installed(raise_exception: Boolean = False) -> Boolean:
    """
    Return whether *OpenImageIO* is installed and available.

    Parameters
    ----------
    raise_exception
        Whether to raise an exception if *OpenImageIO* is unavailable.

    Returns
    -------
    :class:`bool`
        Whether *OpenImageIO* is installed.

    Raises
    ------
    :class:`ImportError`
        If *OpenImageIO* is not installed.
    """

    try:  # pragma: no cover
        # pylint: disable=W0612
        import OpenImageIO  # noqa

        return True
    except ImportError as error:  # pragma: no cover
        if raise_exception:
            raise ImportError(
                '"OpenImageIO" related API features are not available: '
                f'"{error}".\nSee the installation guide for more information: '
                "https://www.colour-science.org/installation-guide/"
            )

        return False


def is_pandas_installed(raise_exception: Boolean = False) -> Boolean:
    """
    Return whether *Pandas* is installed and available.

    Parameters
    ----------
    raise_exception
        Whether to raise an exception if *Pandas* is unavailable.

    Returns
    -------
    :class:`bool`
        Whether *Pandas* is installed.

    Raises
    ------
    :class:`ImportError`
        If *Pandas* is not installed.
    """

    try:  # pragma: no cover
        # pylint: disable=W0612
        import pandas  # noqa

        return True
    except ImportError as error:  # pragma: no cover
        if raise_exception:
            raise ImportError(
                f'"Pandas" related API features are not available: "{error}".\n'
                "See the installation guide for more information: "
                "https://www.colour-science.org/installation-guide/"
            )

        return False


def is_sklearn_installed(raise_exception: Boolean = False) -> Boolean:
    """
    Return whether *Scikit-Learn* (sklearn) is installed and available.

    Parameters
    ----------
    raise_exception
        Whether to raise an exception if *Scikit-Learn* (sklearn) is
        unavailable.

    Returns
    -------
    :class:`bool`
        Whether *Scikit- isLearn* (sklearn) installed.

    Raises
    ------
    :class:`ImportError`
        If *Scikit-Learn* (sklearn) is not installed.
    """

    try:  # pragma: no cover
        # pylint: disable=W0612
        import sklearn  # noqa

        return True
    except ImportError as error:  # pragma: no cover
        if raise_exception:
            raise ImportError(
                '"Scikit-Learn" related API features are not available: '
                f'"{error}".\nSee the installation guide for more information: '
                "https://www.colour-science.org/installation-guide/"
            )

        return False


def is_tqdm_installed(raise_exception: Boolean = False) -> Boolean:
    """
    Return whether *tqdm* is installed and available.

    Parameters
    ----------
    raise_exception
        Whether to raise an exception if *tqdm* is unavailable.

    Returns
    -------
    :class:`bool`
        Whether *tqdm* is installed.

    Raises
    ------
    :class:`ImportError`
        If *tqdm* is not installed.
    """

    try:  # pragma: no cover
        # pylint: disable=W0612
        import tqdm  # noqa

        return True
    except ImportError as error:  # pragma: no cover
        if raise_exception:
            raise ImportError(
                f'"tqdm" related API features are not available: "{error}".\n'
                "See the installation guide for more information: "
                "https://www.colour-science.org/installation-guide/"
            )

        return False


def is_trimesh_installed(raise_exception: Boolean = False) -> Boolean:
    """
    Return whether *Trimesh* is installed and available.

    Parameters
    ----------
    raise_exception
        Whether to raise an exception if *Trimesh* is unavailable.

    Returns
    -------
    :class:`bool`
        Whether *Trimesh* is installed.

    Raises
    ------
    :class:`ImportError`
        If *Trimesh* is not installed.
    """

    try:  # pragma: no cover
        # pylint: disable=W0612
        import trimesh  # noqa

        return True
    except ImportError as error:  # pragma: no cover
        if raise_exception:
            raise ImportError(
                '"Trimesh" related API features are not available: '
                f'"{error}".\nSee the installation guide for more information: '
                "https://www.colour-science.org/installation-guide/"
            )

        return False


_REQUIREMENTS_TO_CALLABLE: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "Matplotlib": is_matplotlib_installed,
        "NetworkX": is_networkx_installed,
        "OpenColorIO": is_opencolorio_installed,
        "OpenImageIO": is_openimageio_installed,
        "Pandas": is_pandas_installed,
        "Scikit-Learn": is_sklearn_installed,
        "tqdm": is_tqdm_installed,
        "trimesh": is_trimesh_installed,
    }
)
"""
Mapping of requirements to their respective callables.

_REQUIREMENTS_TO_CALLABLE
    **{'Matplotlib', 'NetworkX', 'OpenColorIO', 'OpenImageIO', 'Pandas',
    'Scikit-Learn', 'tqdm', 'trimesh'}**
"""


def required(
    *requirements: Literal[
        "Matplotlib",
        "NetworkX",
        "OpenColorIO",
        "OpenImageIO",
        "Pandas",
        "Scikit-Learn",
        "tqdm",
        "trimesh",
    ]
) -> Callable:
    """
    Decorate a function to check whether various ancillary package requirements
    are satisfied.

    Other Parameters
    ----------------
    requirements
        Requirements to check whether they are satisfied.

    Returns
    -------
    Callable
    """

    def wrapper(function: Callable) -> Callable:
        """Wrap given function wrapper."""

        @functools.wraps(function)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            """Wrap given function."""

            for requirement in requirements:
                _REQUIREMENTS_TO_CALLABLE[requirement](raise_exception=True)

            return function(*args, **kwargs)

        return wrapped

    return wrapper


def is_iterable(a: Any) -> Boolean:
    """
    Return whether given variable :math:`a` is iterable.

    Parameters
    ----------
    a
        Variable :math:`a` to check the iterability.

    Returns
    -------
    :class:`bool`
        Whether variable :math:`a` is iterable.

    Examples
    --------
    >>> is_iterable([1, 2, 3])
    True
    >>> is_iterable(1)
    False
    """

    return is_string(a) or (True if getattr(a, "__iter__", False) else False)


def is_string(a: Any) -> Boolean:
    """
    Return whether given variable :math:`a` is a :class:`str`-like variable.

    Parameters
    ----------
    a
        Variable :math:`a` to test.

    Returns
    -------
    :class:`bool`
        Whether variable :math:`a` is a :class:`str`-like variable.

    Examples
    --------
    >>> is_string("I'm a string!")
    True
    >>> is_string(["I'm a string!"])
    False
    """

    return True if isinstance(a, str) else False


def is_numeric(a: Any) -> Boolean:
    """
    Return whether given variable :math:`a` is a :class:`Number`-like
    variable.

    Parameters
    ----------
    a
        Variable :math:`a` to test.

    Returns
    -------
    :class:`bool`
        Whether variable :math:`a` is a :class:`Number`-like variable.

    Examples
    --------
    >>> is_numeric(1)
    True
    >>> is_numeric((1,))
    False
    """

    return isinstance(
        a,
        (
            int,
            float,
            complex,
            np.integer,
            np.floating,
        ),
    )


def is_integer(a: Any) -> Boolean:
    """
    Return whether given variable :math:`a` is an :class:`numpy.integer`-like
    variable under given threshold.

    Parameters
    ----------
    a
        Variable :math:`a` to test.

    Returns
    -------
    :class:`bool`
        Whether variable :math:`a` is an :class:`numpy.integer`-like variable.

    Notes
    -----
    -   The determination threshold is defined by the
        :attr:`colour.algebra.common.INTEGER_THRESHOLD` attribute.

    Examples
    --------
    >>> is_integer(1)
    True
    >>> is_integer(1.01)
    False
    """

    return abs(a - np.around(a)) <= INTEGER_THRESHOLD


def is_sibling(element: Any, mapping: Mapping) -> Boolean:
    """
    Return whether given element type is present in given mapping types.

    Parameters
    ----------
    element
        Element to check whether its type is present in the mapping types.
    mapping
        Mapping types.

    Returns
    -------
    :class:`bool`
        Whether given element type is present in given mapping types.
    """

    return isinstance(
        element, tuple({type(element) for element in mapping.values()})
    )


def filter_kwargs(function: Callable, **kwargs: Any) -> Dict:
    """
    Filter keyword arguments incompatible with the given function signature.

    Parameters
    ----------
    function
        Callable to filter the incompatible keyword arguments.

    Other Parameters
    ----------------
    kwargs
        Keywords arguments.

    Returns
    -------
    dict
        Filtered keyword arguments.

    Examples
    --------
    >>> def fn_a(a):
    ...     return a
    >>> def fn_b(a, b=0):
    ...     return a, b
    >>> def fn_c(a, b=0, c=0):
    ...     return a, b, c
    >>> fn_a(1, **filter_kwargs(fn_a, b=2, c=3))
    1
    >>> fn_b(1, **filter_kwargs(fn_b, b=2, c=3))
    (1, 2)
    >>> fn_c(1, **filter_kwargs(fn_c, b=2, c=3))
    (1, 2, 3)
    """

    kwargs = copy(kwargs)

    try:
        args = list(inspect.signature(function).parameters.keys())
    except ValueError:  # pragma: no cover
        return {}

    for key in set(kwargs.keys()) - set(args):
        kwargs.pop(key)

    return kwargs


def filter_mapping(
    mapping: Mapping,
    filterers: Union[str, Sequence[str]],
    anchors: Boolean = True,
    flags: Union[Integer, RegexFlag] = re.IGNORECASE,
) -> Dict:
    """
    Filter given mapping with given filterers.

    Parameters
    ----------
    mapping
        Mapping to filter.
    filterers
        Filterer pattern for given mapping elements or a list of filterers.
    anchors
        Whether to use Regex line anchors, i.e. *^* and *$* are added,
        surrounding the filterer pattern.
    flags
        Regex flags.

    Returns
    -------
    dict
        Filtered mapping elements.

    Notes
    -----
    -   To honour the filterers ordering, the return value is an
        :class:`dict` class instance.

    Examples
    --------
    >>> class Element:
    ...     pass
    >>> mapping = {
    ...     'Element A': Element(),
    ...     'Element B': Element(),
    ...     'Element C': Element(),
    ...     'Not Element C': Element(),
    ... }
    >>> filter_mapping(mapping, '\\w+\\s+A')  # doctest: +ELLIPSIS
    {'Element A': <colour.utilities.common.Element object at 0x...>}
    >>> sorted(filter_mapping(mapping, 'Element.*'))
    ['Element A', 'Element B', 'Element C']
    """

    def filter_mapping_with_filter(
        mapping: Mapping,
        filterer: str,
        anchors: Boolean = True,
        flags: Union[Integer, RegexFlag] = re.IGNORECASE,
    ) -> Dict:
        """
        Filter given mapping with given filterer.

        Parameters
        ----------
        mapping
            Mapping to filter.
        filterer
            Filterer pattern for given mapping elements.
        anchors
            Whether to use Regex line anchors, i.e. *^* and *$* are added,
            surrounding the filterer pattern.
        flags
            Regex flags.

        Returns
        -------
        dict
            Filtered mapping elements.
        """

        if anchors:
            filterer = f"^{filterer}$"
            filterer = filterer.replace("^^", "^").replace("$$", "$")

        elements = [
            mapping[element]
            for element in mapping
            if re.match(filterer, element, flags)
        ]

        lookup = Lookup(mapping)

        return {
            lookup.first_key_from_value(element): element
            for element in elements
        }

    filterers = [str(filterers)] if is_string(filterers) else filterers

    filtered_mapping = {}

    for filterer in filterers:
        filtered_mapping.update(
            filter_mapping_with_filter(mapping, filterer, anchors, flags)
        )

    return filtered_mapping


def first_item(a: Iterable) -> Any:
    """
    Return the first item of given iterable.

    Parameters
    ----------
    a
        Iterable to get the first item from.

    Returns
    -------
    :class:`object`

    Raises
    ------
    :class:`StopIteration`
        If the iterable is empty.

    Examples
    --------
    >>> a = range(10)
    >>> first_item(a)
    0
    """

    return next(iter(a))


def copy_definition(
    definition: Callable, name: Optional[str] = None
) -> Callable:
    """
    Copy a definition using the same code, globals, defaults, closure, and
    name.

    Parameters
    ----------
    definition
        Definition to be copied.
    name
        Optional definition copy name.

    Returns
    -------
    Callable
        Definition copy.
    """

    copy = types.FunctionType(
        definition.__code__,
        definition.__globals__,  # type: ignore[attr-defined]
        str(name or definition.__name__),
        definition.__defaults__,  # type: ignore[attr-defined]
        definition.__closure__,  # type: ignore[attr-defined]
    )
    copy.__dict__.update(definition.__dict__)

    return copy


def validate_method(
    method: str,
    valid_methods: Union[Sequence, Mapping],
    message: str = '"{0}" method is invalid, it must be one of {1}!',
) -> str:
    """
    Validate whether given method exists in the given valid methods and
    returns the method lower cased.

    Parameters
    ----------
    method
        Method to validate.
    valid_methods
        Valid methods.
    message
        Message for the exception.

    Returns
    -------
    :class:`str`
        Method lower cased.

    Raises
    ------
    :class:`ValueError`
         If the method does not exist.

    Examples
    --------
    >>> validate_method('Valid', ['Valid', 'Yes', 'Ok'])
    'valid'
    """

    valid_methods = [str(valid_method) for valid_method in valid_methods]

    method_lower = method.lower()
    if method_lower not in [
        valid_method.lower() for valid_method in valid_methods
    ]:
        raise ValueError(message.format(method, valid_methods))

    return method_lower


T = TypeVar("T")


def optional(value: Optional[T], default: T) -> T:
    """
    Handle optional argument value by providing a default value.

    Parameters
    ----------
    value
        Optional argument value.
    default
        Default argument value if ``value`` is *None*.

    Returns
    -------
    T
        Argument value.

    Examples
    --------
    >>> optional('Foo', 'Bar')
    'Foo'
    >>> optional(None, 'Bar')
    'Bar'
    """

    if value is None:
        return default
    else:
        return value



"""Defines the unit tests for the :mod:`colour.difference.cam02_ucs` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.difference import (
    delta_E_CAM02LCD,
    delta_E_CAM02SCD,
    delta_E_CAM02UCS,
)
from colour.difference.cam02_ucs import delta_E_Luo2006
from colour.models.cam02_ucs import COEFFICIENTS_UCS_LUO2006
from colour.utilities import ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestDelta_E_Luo2006",
]


class TestDelta_E_Luo2006(unittest.TestCase):
    """
    Define :func:`colour.difference.cam02_ucs.delta_E_Luo2006` definition unit
    tests methods.
    """

    def test_delta_E_Luo2006(self):
        """Test :func:`colour.difference.cam02_ucs.delta_E_Luo2006` definition."""

        self.assertAlmostEqual(
            delta_E_Luo2006(
                np.array([54.90433134, -0.08450395, -0.06854831]),
                np.array([54.80352754, -3.96940084, -13.57591013]),
                COEFFICIENTS_UCS_LUO2006["CAM02-LCD"],
            ),
            14.055546437777583,
            places=7,
        )

        self.assertAlmostEqual(
            delta_E_Luo2006(
                np.array([54.90433134, -0.08450395, -0.06854831]),
                np.array([54.80352754, -3.96940084, -13.57591013]),
                COEFFICIENTS_UCS_LUO2006["CAM02-LCD"],
            ),
            delta_E_CAM02LCD(
                np.array([54.90433134, -0.08450395, -0.06854831]),
                np.array([54.80352754, -3.96940084, -13.57591013]),
            ),
            places=7,
        )

        self.assertAlmostEqual(
            delta_E_Luo2006(
                np.array([54.90433134, -0.08450395, -0.06854831]),
                np.array([54.80352754, -3.96940084, -13.57591013]),
                COEFFICIENTS_UCS_LUO2006["CAM02-SCD"],
            ),
            delta_E_CAM02SCD(
                np.array([54.90433134, -0.08450395, -0.06854831]),
                np.array([54.80352754, -3.96940084, -13.57591013]),
            ),
            places=7,
        )

        self.assertAlmostEqual(
            delta_E_Luo2006(
                np.array([54.90433134, -0.08450395, -0.06854831]),
                np.array([54.80352754, -3.96940084, -13.57591013]),
                COEFFICIENTS_UCS_LUO2006["CAM02-UCS"],
            ),
            delta_E_CAM02UCS(
                np.array([54.90433134, -0.08450395, -0.06854831]),
                np.array([54.80352754, -3.96940084, -13.57591013]),
            ),
            places=7,
        )

    def test_n_dimensional_delta_E_Luo2006(self):
        """
        Test :func:`colour.difference.cam02_ucs.delta_E_Luo2006` definition
        n-dimensional arrays support.
        """

        Jpapbp_1 = np.array([54.90433134, -0.08450395, -0.06854831])
        Jpapbp_2 = np.array([54.80352754, -3.96940084, -13.57591013])
        delta_E_p = delta_E_Luo2006(
            Jpapbp_1, Jpapbp_2, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
        )

        Jpapbp_1 = np.tile(Jpapbp_1, (6, 1))
        Jpapbp_2 = np.tile(Jpapbp_2, (6, 1))
        delta_E_p = np.tile(delta_E_p, 6)
        np.testing.assert_almost_equal(
            delta_E_Luo2006(
                Jpapbp_1, Jpapbp_2, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
            ),
            delta_E_p,
            decimal=7,
        )

        Jpapbp_1 = np.reshape(Jpapbp_1, (2, 3, 3))
        Jpapbp_2 = np.reshape(Jpapbp_2, (2, 3, 3))
        delta_E_p = np.reshape(delta_E_p, (2, 3))
        np.testing.assert_almost_equal(
            delta_E_Luo2006(
                Jpapbp_1, Jpapbp_2, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
            ),
            delta_E_p,
            decimal=7,
        )

    @ignore_numpy_errors
    def test_nan_delta_E_Luo2006(self):
        """
        Test :func:`colour.difference.cam02_ucs.delta_E_Luo2006`
        definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            Jpapbp_1 = np.array(case)
            Jpapbp_2 = np.array(case)
            delta_E_Luo2006(
                Jpapbp_1, Jpapbp_2, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
            ),


if __name__ == "__main__":
    unittest.main()



"""Showcases corresponding chromaticities prediction computations."""

from pprint import pprint

import colour
from colour.utilities import message_box

message_box("Corresponding Chromaticities Prediction Computations")

message_box(
    'Computing corresponding chromaticities prediction with "Von Kries" '
    'chromatic adaptation model for "Breneman (1987)" experiment number "3" '
    'and "Bianco" CAT.'
)
pprint(
    colour.corresponding_chromaticities_prediction(
        3, model="Von Kries", transform="Bianco 2010"
    )
)
pprint(
    colour.corresponding.corresponding_chromaticities_prediction_VonKries(
        3, "Bianco 2010"
    )
)

print("\n")

message_box(
    'Computing corresponding chromaticities prediction with "CIE 1994" '
    'chromatic adaptation model for "Breneman (1987)" experiment number "1".'
)
pprint(colour.corresponding_chromaticities_prediction(3, model="CIE 1994"))
pprint(colour.corresponding.corresponding_chromaticities_prediction_CIE1994(1))

print("\n")

message_box(
    'Computing corresponding chromaticities prediction with "CMCCAT2000" '
    'chromatic adaptation model for "Breneman (1987)" experiment number "1".'
)
pprint(colour.corresponding_chromaticities_prediction(3, model="CMCCAT2000"))
pprint(
    colour.corresponding.corresponding_chromaticities_prediction_CMCCAT2000(1)
)

print("\n")

message_box(
    'Computing corresponding chromaticities prediction with Fairchild (1990)" '
    'chromatic adaptation model for "Breneman (1987)" experiment number "1".'
)
pprint(
    colour.corresponding_chromaticities_prediction(3, model="Fairchild 1990")
)
pprint(
    colour.corresponding.corresponding_chromaticities_prediction_Fairchild1990(
        1
    )
)



"""
References
----------
-   :cite:`AdobeSystems2013` : Adobe Systems. (2013). Adobe DNG Software
    Development Kit (SDK) - 1.3.0.0 -
    dng_sdk_1_3/dng_sdk/source/dng_temperature.cpp::dng_temperature::\
Set_xy_coord. https://www.adobe.com/support/downloads/dng/dng_sdk.html
-   :cite:`AdobeSystems2013a` : Adobe Systems. (2013). Adobe DNG Software
    Development Kit (SDK) - 1.3.0.0 -
    dng_sdk_1_3/dng_sdk/source/dng_temperature.cpp::dng_temperature::xy_coord.
    https://www.adobe.com/support/downloads/dng/dng_sdk.html
-   :cite:`Hernandez-Andres1999a` : Hernández-Andrés, J., Lee, R. L., &
    Romero, J. (1999). Calculating correlated color temperatures across the
    entire gamut of daylight and skylight chromaticities. Applied Optics,
    38(27),
    5703. doi:10.1364/AO.38.005703
-   :cite:`Kang2002a` : Kang, B., Moon, O., Hong, C., Lee, H., Cho, B., & Kim,
    Y. (2002). Design of advanced color: Temperature control system for HDTV
    applications. Journal of the Korean Physical Society, 41(6), 865-871.
-   :cite:`Krystek1985b` : Krystek, M. (1985). An algorithm to calculate
    correlated colour temperature. Color Research & Application, 10(1), 38-40.
    doi:10.1002/col.5080100109
-   :cite:`Ohno2014a` : Ohno, Yoshiro. (2014). Practical Use and Calculation of
    CCT and Duv. LEUKOS, 10(1), 47-55. doi:10.1080/15502724.2014.839020
-   :cite:`Wikipedia2001` : Wikipedia. (2001). Approximation. Retrieved June
    28, 2014, from http://en.wikipedia.org/wiki/Color_temperature#Approximation
-   :cite:`Wikipedia2001a` : Wikipedia. (2001). Color temperature. Retrieved
    June 28, 2014, from http://en.wikipedia.org/wiki/Color_temperature
-   :cite:`Wyszecki2000y` : Wyszecki, Günther, & Stiles, W. S. (2000).
    DISTRIBUTION TEMPERATURE, COLOR TEMPERATURE, AND CORRELATED COLOR
    TEMPERATURE. In Color Science: Concepts and Methods, Quantitative Data and
    Formulae (pp. 224-229). Wiley. ISBN:978-0-471-39918-6
-   :cite:`Wyszecki2000z` : Wyszecki, Günther, & Stiles, W. S. (2000). CIE
    Method of Calculating D-Illuminants. In Color Science: Concepts and
    Methods, Quantitative Data and Formulae (pp. 145-146). Wiley.
    ISBN:978-0-471-39918-6
"""

from __future__ import annotations

from colour.hints import (
    Any,
    ArrayLike,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    NDArray,
    Literal,
    Union,
)
from colour.utilities import (
    CaseInsensitiveMapping,
    filter_kwargs,
    validate_method,
)

from .cie_d import xy_to_CCT_CIE_D, CCT_to_xy_CIE_D
from .hernandez1999 import xy_to_CCT_Hernandez1999, CCT_to_xy_Hernandez1999
from .kang2002 import xy_to_CCT_Kang2002, CCT_to_xy_Kang2002
from .krystek1985 import uv_to_CCT_Krystek1985, CCT_to_uv_Krystek1985
from .mccamy1992 import xy_to_CCT_McCamy1992, CCT_to_xy_McCamy1992
from .ohno2013 import uv_to_CCT_Ohno2013, CCT_to_uv_Ohno2013
from .robertson1968 import uv_to_CCT_Robertson1968, CCT_to_uv_Robertson1968

__all__ = [
    "xy_to_CCT_CIE_D",
    "CCT_to_xy_CIE_D",
]
__all__ += [
    "xy_to_CCT_Hernandez1999",
    "CCT_to_xy_Hernandez1999",
]
__all__ += [
    "xy_to_CCT_Kang2002",
    "CCT_to_xy_Kang2002",
]
__all__ += [
    "uv_to_CCT_Krystek1985",
    "CCT_to_uv_Krystek1985",
]
__all__ += [
    "xy_to_CCT_McCamy1992",
    "CCT_to_xy_McCamy1992",
]
__all__ += [
    "uv_to_CCT_Ohno2013",
    "CCT_to_uv_Ohno2013",
]
__all__ += [
    "uv_to_CCT_Robertson1968",
    "CCT_to_uv_Robertson1968",
]

UV_TO_CCT_METHODS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "Krystek 1985": uv_to_CCT_Krystek1985,
        "Ohno 2013": uv_to_CCT_Ohno2013,
        "Robertson 1968": uv_to_CCT_Robertson1968,
    }
)
UV_TO_CCT_METHODS.__doc__ = """
Supported *CIE UCS* colourspace *uv* chromaticity coordinates to correlated
colour temperature :math:`T_{cp}` computation methods.

References
----------
:cite:`AdobeSystems2013`, :cite:`AdobeSystems2013a`, :cite:`Krystek1985b`,
:cite:`Ohno2014a`, :cite:`Wyszecki2000y`

Aliases:

-   'ohno2013': 'Ohno 2013'
-   'robertson1968': 'Robertson 1968'
"""
UV_TO_CCT_METHODS["ohno2013"] = UV_TO_CCT_METHODS["Ohno 2013"]
UV_TO_CCT_METHODS["robertson1968"] = UV_TO_CCT_METHODS["Robertson 1968"]


def uv_to_CCT(
    uv: ArrayLike,
    method: Union[
        Literal["Krystek 1985", "Ohno 2013", "Robertson 1968"], str
    ] = "Ohno 2013",
    **kwargs: Any,
) -> NDArray:
    """
    Return the correlated colour temperature :math:`T_{cp}` and
    :math:`\\Delta_{uv}` from given *CIE UCS* colourspace *uv* chromaticity
    coordinates using given method.

    Parameters
    ----------
    uv
        *CIE UCS* colourspace *uv* chromaticity coordinates.
    method
        Computation method.

    Other Parameters
    ----------------
    cmfs
        {:func:`colour.temperature.uv_to_CCT_Ohno2013`},
        Standard observer colour matching functions.
    count
        {:func:`colour.temperature.uv_to_CCT_Ohno2013`},
        Temperatures count in the planckian tables.
    end
        {:func:`colour.temperature.uv_to_CCT_Ohno2013`},
        Temperature range end in kelvins.
    iterations
        {:func:`colour.temperature.uv_to_CCT_Ohno2013`},
        Number of planckian tables to generate.
    optimisation_kwargs
        {:func:`colour.temperature.uv_to_CCT_Krystek1985`},
        Parameters for :func:`scipy.optimize.minimize` definition.
    start
        {:func:`colour.temperature.uv_to_CCT_Ohno2013`},
        Temperature range start in kelvins.

    Returns
    -------
    :class:`numpy.ndarray`
        Correlated colour temperature :math:`T_{cp}`, :math:`\\Delta_{uv}`.

    References
    ----------
    :cite:`AdobeSystems2013`, :cite:`AdobeSystems2013a`, :cite:`Krystek1985b`,
    :cite:`Ohno2014a`, :cite:`Wyszecki2000y`

    Examples
    --------
    >>> import numpy as np
    >>> uv = np.array([0.1978, 0.3122])
    >>> uv_to_CCT(uv)  # doctest: +ELLIPSIS
    array([  6.507473...e+03,   3.223346...e-03])
    """

    method = validate_method(method, UV_TO_CCT_METHODS)

    function = UV_TO_CCT_METHODS[method]

    return function(uv, **filter_kwargs(function, **kwargs))


CCT_TO_UV_METHODS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "Krystek 1985": CCT_to_uv_Krystek1985,
        "Ohno 2013": CCT_to_uv_Ohno2013,
        "Robertson 1968": CCT_to_uv_Robertson1968,
    }
)
CCT_TO_UV_METHODS.__doc__ = """
Supported correlated colour temperature :math:`T_{cp}` to *CIE UCS* colourspace
*uv* chromaticity coordinates computation methods.

References
----------
:cite:`AdobeSystems2013`, :cite:`AdobeSystems2013a`, :cite:`Krystek1985b`,
:cite:`Ohno2014a`, :cite:`Wyszecki2000y`

Aliases:

-   'ohno2013': 'Ohno 2013'
-   'robertson1968': 'Robertson 1968'
"""
CCT_TO_UV_METHODS["ohno2013"] = CCT_TO_UV_METHODS["Ohno 2013"]
CCT_TO_UV_METHODS["robertson1968"] = CCT_TO_UV_METHODS["Robertson 1968"]


def CCT_to_uv(
    CCT_D_uv: ArrayLike,
    method: Union[
        Literal["Krystek 1985", "Ohno 2013", "Robertson 1968"], str
    ] = "Ohno 2013",
    **kwargs: Any,
) -> NDArray:
    """
    Return the *CIE UCS* colourspace *uv* chromaticity coordinates from given
    correlated colour temperature :math:`T_{cp}` using given method.

    Parameters
    ----------
    CCT_D_uv
        Correlated colour temperature :math:`T_{cp}`, :math:`\\Delta_{uv}`.
    method
        Computation method.

    Other Parameters
    ----------------
    cmfs
        {:func:`colour.temperature.CCT_to_uv_Ohno2013`},
        Standard observer colour matching functions.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE UCS* colourspace *uv* chromaticity coordinates.

    References
    ----------
    :cite:`AdobeSystems2013`, :cite:`AdobeSystems2013a`, :cite:`Krystek1985b`,
    :cite:`Ohno2014a`, :cite:`Wyszecki2000y`

    Examples
    --------
    >>> import numpy as np
    >>> CCT_D_uv = np.array([6507.47380460, 0.00322335])
    >>> CCT_to_uv(CCT_D_uv)  # doctest: +ELLIPSIS
    array([ 0.1977999...,  0.3121999...])
    """

    method = validate_method(method, CCT_TO_UV_METHODS)

    function = CCT_TO_UV_METHODS[method]

    return function(CCT_D_uv, **filter_kwargs(function, **kwargs))


__all__ += [
    "UV_TO_CCT_METHODS",
    "uv_to_CCT",
]
__all__ += [
    "CCT_TO_UV_METHODS",
    "CCT_to_uv",
]

XY_TO_CCT_METHODS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "CIE Illuminant D Series": xy_to_CCT_CIE_D,
        "Hernandez 1999": xy_to_CCT_Hernandez1999,
        "Kang 2002": xy_to_CCT_Kang2002,
        "McCamy 1992": xy_to_CCT_McCamy1992,
    }
)
XY_TO_CCT_METHODS.__doc__ = """
Supported *CIE xy* chromaticity coordinates to correlated colour temperature
:math:`T_{cp}` computation methods.

References
----------
:cite:`Hernandez-Andres1999a`, :cite:`Kang2002a`, :cite:`Wikipedia2001`,
:cite:`Wikipedia2001a`, :cite:`Wyszecki2000z`

Aliases:

-   'daylight': 'CIE Illuminant D Series'
-   'kang2002': 'Kang 2002'
-   'mccamy1992': 'McCamy 1992'
-   'hernandez1999': 'Hernandez 1999'
"""
XY_TO_CCT_METHODS["daylight"] = XY_TO_CCT_METHODS["CIE Illuminant D Series"]
XY_TO_CCT_METHODS["kang2002"] = XY_TO_CCT_METHODS["Kang 2002"]
XY_TO_CCT_METHODS["mccamy1992"] = XY_TO_CCT_METHODS["McCamy 1992"]
XY_TO_CCT_METHODS["hernandez1999"] = XY_TO_CCT_METHODS["Hernandez 1999"]


def xy_to_CCT(
    xy: ArrayLike,
    method: Union[
        Literal[
            "CIE Illuminant D Series",
            "Kang 2002",
            "Hernandez 1999",
            "McCamy 1992",
        ],
        str,
    ] = "CIE Illuminant D Series",
) -> FloatingOrNDArray:
    """
    Return the correlated colour temperature :math:`T_{cp}` from given
    *CIE xy* chromaticity coordinates using given method.

    Parameters
    ----------
    xy
        *CIE xy* chromaticity coordinates.
    method
        Computation method.

    Other Parameters
    ----------------
    optimisation_kwargs
        {:func:`colour.temperature.xy_to_CCT_CIE_D`,
        :func:`colour.temperature.xy_to_CCT_Kang2002`},
        Parameters for :func:`scipy.optimize.minimize` definition.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Correlated colour temperature :math:`T_{cp}`.

    References
    ----------
    :cite:`Hernandez-Andres1999a`, :cite:`Kang2002a`, :cite:`Wikipedia2001`,
    :cite:`Wikipedia2001a`, :cite:`Wyszecki2000z`

    Examples
    --------
    >>> import numpy as np
    >>> xy_to_CCT(np.array([0.31270, 0.32900]))  # doctest: +ELLIPSIS
    6508.1175148...
    >>> xy_to_CCT(np.array([0.31270, 0.32900]), 'Hernandez 1999')
    ... # doctest: +ELLIPSIS
    6500.7420431...
    """

    method = validate_method(method, XY_TO_CCT_METHODS)

    return XY_TO_CCT_METHODS[method](xy)


CCT_TO_XY_METHODS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "CIE Illuminant D Series": CCT_to_xy_CIE_D,
        "Hernandez 1999": CCT_to_xy_Hernandez1999,
        "Kang 2002": CCT_to_xy_Kang2002,
        "McCamy 1992": CCT_to_xy_McCamy1992,
    }
)
CCT_TO_XY_METHODS.__doc__ = """
Supported correlated colour temperature :math:`T_{cp}` to *CIE xy* chromaticity
coordinates computation methods.

References
----------
:cite:`Hernandez-Andres1999a`, :cite:`Kang2002a`, :cite:`Wikipedia2001`,
:cite:`Wikipedia2001a`, :cite:`Wyszecki2000z`

Aliases:

-   'daylight': 'CIE Illuminant D Series'
-   'kang2002': 'Kang 2002'
-   'mccamy1992': 'McCamy 1992'
-   'hernandez1999': 'Hernandez 1999'
"""
CCT_TO_XY_METHODS["daylight"] = CCT_TO_XY_METHODS["CIE Illuminant D Series"]
CCT_TO_XY_METHODS["kang2002"] = CCT_TO_XY_METHODS["Kang 2002"]
CCT_TO_XY_METHODS["mccamy1992"] = CCT_TO_XY_METHODS["McCamy 1992"]
CCT_TO_XY_METHODS["hernandez1999"] = CCT_TO_XY_METHODS["Hernandez 1999"]


def CCT_to_xy(
    CCT: FloatingOrArrayLike,
    method: Union[
        Literal[
            "CIE Illuminant D Series",
            "Kang 2002",
            "Hernandez 1999",
            "McCamy 1992",
        ],
        str,
    ] = "CIE Illuminant D Series",
) -> NDArray:
    """
    Return the *CIE xy* chromaticity coordinates from given correlated colour
    temperature :math:`T_{cp}` using given method.

    Parameters
    ----------
    CCT
        Correlated colour temperature :math:`T_{cp}`.
    method
        Computation method.

    Other Parameters
    ----------------
    optimisation_kwargs
        {:func:`colour.temperature.CCT_to_xy_Hernandez1999`,
        :func:`colour.temperature.CCT_to_xy_McCamy1992`},
        Parameters for :func:`scipy.optimize.minimize` definition.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE xy* chromaticity coordinates.

    References
    ----------
    :cite:`Hernandez-Andres1999a`, :cite:`Kang2002a`, :cite:`Wikipedia2001`,
    :cite:`Wikipedia2001a`, :cite:`Wyszecki2000z`

    Examples
    --------
    >>> CCT_to_xy(6504.38938305)  # doctest: +ELLIPSIS
    array([ 0.3127077...,  0.3291128...])
    >>> CCT_to_xy(6504.38938305, 'Kang 2002')
    ... # doctest: +ELLIPSIS
    array([ 0.313426 ...,  0.3235959...])
    """

    method = validate_method(method, CCT_TO_XY_METHODS)

    return CCT_TO_XY_METHODS[method](CCT)


__all__ += [
    "XY_TO_CCT_METHODS",
    "xy_to_CCT",
]
__all__ += [
    "CCT_TO_XY_METHODS",
    "CCT_to_xy",
]



"""
Linear Colour Component Transfer Function
=========================================

Defines the linear encoding / decoding colour component transfer function
related objects:

- :func:`colour.linear_function`
"""

from __future__ import annotations

from colour.hints import FloatingOrArrayLike, FloatingOrNDArray
from colour.utilities import as_float

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "linear_function",
]


def linear_function(a: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Define a typical linear encoding / decoding function, essentially a
    pass-through function.

    Parameters
    ----------
    a
        Array to encode / decode.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Encoded / decoded array.

    Examples
    --------
    >>> linear_function(0.18)  # doctest: +ELLIPSIS
    0.1799999...
    """

    return as_float(a)



"""Showcases colour quality plotting examples."""

import colour
from colour.plotting import (
    colour_style,
    plot_multi_sds_colour_quality_scales_bars,
    plot_multi_sds_colour_rendering_indexes_bars,
    plot_single_sd_colour_quality_scale_bars,
    plot_single_sd_colour_rendering_index_bars,
)
from colour.utilities import message_box

message_box("Colour Quality Plots")

colour_style()

message_box('Plotting "F2" illuminant "Colour Rendering Index (CRI)".')
plot_single_sd_colour_rendering_index_bars(colour.SDS_ILLUMINANTS["FL2"])

print("\n")

message_box(
    "Plotting various illuminants and light sources "
    '"Colour Rendering Index (CRI)".'
)
plot_multi_sds_colour_rendering_indexes_bars(
    (
        colour.SDS_ILLUMINANTS["FL2"],
        colour.SDS_LIGHT_SOURCES["F32T8/TL841 (Triphosphor)"],
        colour.SDS_LIGHT_SOURCES["Kinoton 75P"],
    )
)

print("\n")

message_box('Plotting "F2" illuminant "Colour Quality Scale (CQS)".')
plot_single_sd_colour_quality_scale_bars(colour.SDS_ILLUMINANTS["FL2"])

print("\n")

message_box(
    "Plotting various illuminants and light sources "
    '"Colour Quality Scale (CQS)".'
)
plot_multi_sds_colour_quality_scales_bars(
    (
        colour.SDS_ILLUMINANTS["FL2"],
        colour.SDS_LIGHT_SOURCES["F32T8/TL841 (Triphosphor)"],
        colour.SDS_LIGHT_SOURCES["Kinoton 75P"],
    )
)



"""
Best RGB Colourspace
====================

Defines the *Best RGB* colourspace:

-   :attr:`colour.models.RGB_COLOURSPACE_BEST_RGB`.

References
----------
-   :cite:`HutchColord` : HutchColor. (n.d.). BestRGB (4 K).
    http://www.hutchcolor.com/profiles/BestRGB.zip
"""

from __future__ import annotations

import numpy as np
from functools import partial

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import NDArray
from colour.models.rgb import (
    RGB_Colourspace,
    gamma_function,
    normalised_primary_matrix,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "PRIMARIES_BEST_RGB",
    "WHITEPOINT_NAME_BEST_RGB",
    "CCS_WHITEPOINT_BEST_RGB",
    "MATRIX_BEST_RGB_TO_XYZ",
    "MATRIX_XYZ_TO_BEST_RGB",
    "RGB_COLOURSPACE_BEST_RGB",
]

PRIMARIES_BEST_RGB: NDArray = np.array(
    [
        [0.735191637630662, 0.264808362369338],
        [0.215336134453781, 0.774159663865546],
        [0.130122950819672, 0.034836065573770],
    ]
)
"""*Best RGB* colourspace primaries."""

WHITEPOINT_NAME_BEST_RGB: str = "D50"
"""*Best RGB* colourspace whitepoint name."""

CCS_WHITEPOINT_BEST_RGB: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
][WHITEPOINT_NAME_BEST_RGB]
"""*Best RGB* colourspace whitepoint chromaticity coordinates."""

MATRIX_BEST_RGB_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_BEST_RGB, CCS_WHITEPOINT_BEST_RGB
)
"""*Best RGB* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_BEST_RGB: NDArray = np.linalg.inv(MATRIX_BEST_RGB_TO_XYZ)
"""*CIE XYZ* tristimulus values to *Best RGB* colourspace matrix."""

RGB_COLOURSPACE_BEST_RGB: RGB_Colourspace = RGB_Colourspace(
    "Best RGB",
    PRIMARIES_BEST_RGB,
    CCS_WHITEPOINT_BEST_RGB,
    WHITEPOINT_NAME_BEST_RGB,
    MATRIX_BEST_RGB_TO_XYZ,
    MATRIX_XYZ_TO_BEST_RGB,
    partial(gamma_function, exponent=1 / 2.2),
    partial(gamma_function, exponent=2.2),
)
RGB_COLOURSPACE_BEST_RGB.__doc__ = """
*Best RGB* colourspace.

References
----------
:cite:`HutchColord`
"""



"""Showcases interpolation computations."""

import matplotlib.pyplot as plt
import numpy as np
import os

import colour
from colour.plotting import render
from colour.utilities import message_box

message_box("Interpolation Computations")

message_box(
    'Comparing the "Sprague (1880)" and "Cubic Spline" recommended '
    'interpolation methods to the "Pchip" method.'
)

data_uniform = {
    340: 0.0000,
    360: 0.0000,
    380: 0.0000,
    400: 0.0641,
    420: 0.0645,
    440: 0.0562,
    460: 0.0537,
    480: 0.0559,
    500: 0.0651,
    520: 0.0705,
    540: 0.0772,
    560: 0.0870,
    580: 0.1128,
    600: 0.1360,
    620: 0.1511,
    640: 0.1688,
    660: 0.1996,
    680: 0.2397,
    700: 0.2852,
    720: 0.0000,
    740: 0.0000,
    760: 0.0000,
    780: 0.0000,
    800: 0.0000,
    820: 0.0000,
}

data_non_uniform = {
    340.1: 0.0000,
    360: 0.0000,
    380: 0.0000,
    400: 0.0641,
    420: 0.0645,
    440: 0.0562,
    460: 0.0537,
    480: 0.0559,
    500: 0.0651,
    520: 0.0705,
    540: 0.0772,
    560: 0.0870,
    580: 0.1128,
    600: 0.1360,
    620: 0.1511,
    640: 0.1688,
    660: 0.1996,
    680: 0.2397,
    700: 0.2852,
    720: 0.0000,
    740: 0.0000,
    760: 0.0000,
    780: 0.0000,
    800: 0.0000,
    820.9: 0.0000,
}

sd_base = colour.SpectralDistribution(data_uniform, name="Reference")
uniform_interpolated_sd = colour.SpectralDistribution(
    data_uniform, name="Uniform - Sprague Interpolation"
)
uniform_pchip_interpolated_sd = colour.SpectralDistribution(
    data_uniform, name="Uniform - Pchip Interpolation"
)
non_uniform_interpolated_sd = colour.SpectralDistribution(
    data_non_uniform, name="Non Uniform - Cubic Spline Interpolation"
)

uniform_interpolated_sd.interpolate(
    colour.SpectralShape(
        uniform_interpolated_sd.shape.start,
        uniform_interpolated_sd.shape.end,
        1,
    )
)
uniform_pchip_interpolated_sd.interpolate(
    colour.SpectralShape(
        uniform_pchip_interpolated_sd.shape.start,
        uniform_pchip_interpolated_sd.shape.end,
        1,
    ),
    interpolator=colour.PchipInterpolator,
)
non_uniform_interpolated_sd.interpolate(
    colour.SpectralShape(
        non_uniform_interpolated_sd.shape.start,
        non_uniform_interpolated_sd.shape.end,
        1,
    )
)

shape = sd_base.shape
x_limit_min, x_limit_max, y_limit_min, y_limit_max = [], [], [], []

plt.plot(
    sd_base.wavelengths, sd_base.values, "ro-", label=sd_base.name, linewidth=1
)
plt.plot(
    uniform_interpolated_sd.wavelengths,
    uniform_interpolated_sd.values,
    label=uniform_interpolated_sd.name,
    linewidth=1,
)
plt.plot(
    uniform_pchip_interpolated_sd.wavelengths,
    uniform_pchip_interpolated_sd.values,
    label=uniform_pchip_interpolated_sd.name,
    linewidth=1,
)
plt.plot(
    non_uniform_interpolated_sd.wavelengths,
    non_uniform_interpolated_sd.values,
    label=non_uniform_interpolated_sd.name,
    linewidth=1,
)

x_limit_min.append(shape.start)
x_limit_max.append(shape.end)
y_limit_min.append(min(sd_base.values))
y_limit_max.append(max(sd_base.values))

settings = {
    "x_label": "Wavelength $\\lambda$ (nm)",
    "y_label": "Spectral Distribution",
    "legend": True,
    "legend_location": "upper left",
    "x_ticker": True,
    "y_ticker": True,
    "bounding_box": (
        min(x_limit_min),
        max(x_limit_max),
        min(y_limit_min),
        max(y_limit_max),
    ),
}

render(**settings)

print("\n")

V_xyz = np.random.random((6, 3))
message_box(
    f'Performing "trilinear" interpolation of given "xyz" values:\n\n'
    f"{V_xyz}\n\n"
    f"using given interpolation table."
)
path = os.path.join(
    os.path.dirname(__file__),
    "..",
    "..",
    "io",
    "luts",
    "tests",
    "resources",
    "iridas_cube",
    "Colour_Correct.cube",
)
table = colour.read_LUT(path).table
print(colour.table_interpolation(V_xyz, table, method="Trilinear"))
print(colour.algebra.table_interpolation_trilinear(V_xyz, table))

print("\n")

message_box(
    f'Performing "tetrahedral" interpolation of given "xyz" values:\n\n'
    f"{V_xyz}\n\n"
    f"using given interpolation table."
)
print(colour.table_interpolation(V_xyz, table, method="Tetrahedral"))
print(colour.algebra.table_interpolation_tetrahedral(V_xyz, table))



"""
:math:`\\Delta E'` - Delta E Colour Difference - Li et al. (2017)
=================================================================

Defines the :math:`\\Delta E'` colour difference computation objects based on
*Li, Li, Wang, Zu, Luo, Cui, Melgosa, Brill and Pointer (2017)* *CAM16-LCD*,
*CAM16-SCD*, and *CAM16-UCS* colourspaces:

-   :func:`colour.difference.delta_E_CAM16LCD`
-   :func:`colour.difference.delta_E_CAM16SCD`
-   :func:`colour.difference.delta_E_CAM16UCS`

References
----------
-   :cite:`Li2017` : Li, C., Li, Z., Wang, Z., Xu, Y., Luo, M. R., Cui, G.,
    Melgosa, M., Brill, M. H., & Pointer, M. (2017). Comprehensive color
    solutions: CAM16, CAT16, and CAM16-UCS. Color Research & Application,
    42(6), 703-718. doi:10.1002/col.22131
"""

from colour.difference.cam02_ucs import (
    delta_E_Luo2006,
    delta_E_CAM02LCD,
    delta_E_CAM02SCD,
    delta_E_CAM02UCS,
)
from colour.models.cam16_ucs import (
    _UCS_Luo2006_callable_to_UCS_Li2017_docstring,
)
from colour.utilities import copy_definition

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "delta_E_Li2017",
    "delta_E_CAM16LCD",
    "delta_E_CAM16SCD",
    "delta_E_CAM16UCS",
]

delta_E_Li2017 = copy_definition(delta_E_Luo2006, "delta_E_Li2017")
delta_E_Li2017.__doc__ = _UCS_Luo2006_callable_to_UCS_Li2017_docstring(
    delta_E_Luo2006
)

delta_E_CAM16LCD = copy_definition(delta_E_CAM02LCD, "delta_E_CAM16LCD")
delta_E_CAM16LCD.__doc__ = _UCS_Luo2006_callable_to_UCS_Li2017_docstring(
    delta_E_CAM02LCD
)

delta_E_CAM16SCD = copy_definition(delta_E_CAM02SCD, "delta_E_CAM16SCD")
delta_E_CAM16SCD.__doc__ = _UCS_Luo2006_callable_to_UCS_Li2017_docstring(
    delta_E_CAM02SCD
)

delta_E_CAM16UCS = copy_definition(delta_E_CAM02UCS, "delta_E_CAM16UCS")
delta_E_CAM16UCS.__doc__ = _UCS_Luo2006_callable_to_UCS_Li2017_docstring(
    delta_E_CAM02UCS
)



"""Showcases Look Up Table (LUT) data related examples."""

import numpy as np
import os

import colour
from colour.utilities import message_box

RESOURCES_DIRECTORY = os.path.join(
    os.path.dirname(__file__), "..", "..", "io", "luts", "tests", "resources"
)

message_box("Look Up Table (LUT) Data")

message_box('Reading a "Cinespace" ".csp" 3D LUT with Shaper file.')
path = os.path.join(
    RESOURCES_DIRECTORY, "cinespace", "Three_Dimensional_Table_With_Shaper.csp"
)
print(colour.io.read_LUT_Cinespace(path))
print("\n")
print(colour.read_LUT(path))

print("\n")

message_box('Reading an "Iridas" ".cube" 3x1D LUT file.')
path = os.path.join(
    RESOURCES_DIRECTORY, "iridas_cube", "ACES_Proxy_10_to_ACES.cube"
)
print(colour.io.read_LUT_IridasCube(path))
print("\n")
print(colour.read_LUT(path))

print("\n")

message_box('Reading an "Iridas" ".cube" 3D LUT file.')
path = os.path.join(RESOURCES_DIRECTORY, "iridas_cube", "Colour_Correct.cube")
print(colour.io.read_LUT_IridasCube(path))
print("\n")
print(colour.read_LUT(path))

print("\n")

message_box('Reading a "Sony" ".spi1d" 1D LUT file.')
path = os.path.join(RESOURCES_DIRECTORY, "sony_spi1d", "eotf_sRGB_1D.spi1d")
print(colour.io.read_LUT_SonySPI1D(path))
print("\n")
print(colour.read_LUT(path))

print("\n")

message_box('Reading a "Sony" ".spi1d" 3x1D LUT file.')
path = os.path.join(RESOURCES_DIRECTORY, "sony_spi1d", "eotf_sRGB_3x1D.spi1d")
print(colour.io.read_LUT_SonySPI1D(path))
print("\n")
print(colour.read_LUT(path))

print("\n")

message_box('Reading a "Sony" ".spi3d" 3D LUT file.')
path = os.path.join(RESOURCES_DIRECTORY, "sony_spi3d", "Colour_Correct.spi3d")
print(colour.io.read_LUT_SonySPI3D(path))
print("\n")
print(colour.read_LUT(path))

print("\n")

message_box('Reading a "Sony" ".spimtx" LUT file.')
path = os.path.join(RESOURCES_DIRECTORY, "sony_spimtx", "dt.spimtx")
print(colour.io.read_LUT_SonySPImtx(path))
print("\n")
print(colour.read_LUT(path))

print("\n")

RGB = np.array([0.35521588, 0.41000000, 0.24177934])
message_box(f'Applying a 1D LUT to given "RGB" values:\n\n\t{RGB}')
path = os.path.join(RESOURCES_DIRECTORY, "sony_spi1d", "eotf_sRGB_1D.spi1d")
LUT = colour.io.read_LUT(path)
print(LUT.apply(RGB))

print("\n")

message_box(f'Applying a 3x1D LUT to given "RGB" values:\n\n\t{RGB}')
path = os.path.join(
    RESOURCES_DIRECTORY, "iridas_cube", "ACES_Proxy_10_to_ACES.cube"
)
LUT = colour.io.read_LUT(path)
print(LUT.apply(RGB))

print("\n")

message_box(f'Applying a 3D LUT to given "RGB" values:\n\n\t{RGB}')
path = os.path.join(RESOURCES_DIRECTORY, "iridas_cube", "Colour_Correct.cube")
LUT = colour.io.read_LUT(path)
print(LUT.apply(RGB))

message_box(
    f'Applying a "Sony" ".spimtx" LUT to given "RGB" values:\n\n\t{RGB}'
)
path = os.path.join(RESOURCES_DIRECTORY, "sony_spimtx", "dt.spimtx")
LUT = colour.io.read_LUT(path)
print(LUT.apply(RGB))



"""
Automatic Colour Conversion Graph Plotting
==========================================

Defines the automatic colour conversion graph plotting objects:

-   :func:`colour.plotting.plot_automatic_colour_conversion_graph`
"""

from __future__ import annotations

import colour
from colour.graph import (
    CONVERSION_GRAPH_NODE_LABELS,
    describe_conversion_path,
)
from colour.hints import Literal, Union
from colour.utilities import required, validate_method

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "plot_automatic_colour_conversion_graph",
]


@required("NetworkX")
def plot_automatic_colour_conversion_graph(
    filename: str,
    prog: Union[
        Literal["circo", "dot", "fdp", "neato", "nop", "twopi"], str
    ] = "fdp",
    args: str = "",
) -> AGraph:  # type: ignore[name-defined]  # noqa
    """
    Plot *Colour* automatic colour conversion graph using
    `Graphviz <https://www.graphviz.org/>`__ and
    `pyraphviz <https://pygraphviz.github.io>`__.

    Parameters
    ----------
    filename
        Filename to use to save the image.
    prog
        *Graphviz* layout method.
    args
         Additional arguments for *Graphviz*.

    Returns
    -------
    :class:`AGraph`
        *Pyraphviz* graph.

    Notes
    -----
    -   This definition does not directly plot the *Colour* automatic colour
        conversion graph but instead write it to an image.

    Examples
    --------
    >>> import tempfile
    >>> import colour
    >>> from colour import read_image
    >>> from colour.plotting import plot_image
    >>> filename = '{0}.png'.format(tempfile.mkstemp()[-1])
    >>> _ = plot_automatic_colour_conversion_graph(filename, 'dot')
    ... # doctest: +SKIP
    >>> plot_image(read_image(filename))  # doctest: +SKIP

    .. image:: ../_static/Plotting_Plot_Colour_Automatic_Conversion_Graph.png
        :align: center
        :alt: plot_automatic_colour_conversion_graph
    """

    import networkx as nx

    prog = validate_method(
        prog,
        ["circo", "dot", "fdp", "neato", "nop", "twopi"],
        '"{0}" program is invalid, it must be one of {1}!',
    )

    # TODO: Investigate API to trigger the conversion graph build.
    describe_conversion_path("RGB", "RGB", print_callable=lambda x: x)

    agraph = nx.nx_agraph.to_agraph(colour.graph.CONVERSION_GRAPH)

    for node in agraph.nodes():
        node.attr.update(label=CONVERSION_GRAPH_NODE_LABELS[node.name])

    agraph.node_attr.update(
        style="filled",
        shape="circle",
        color="#2196F3FF",
        fillcolor="#2196F370",
        fontname="Helvetica",
        fontcolor="#263238",
    )
    agraph.edge_attr.update(color="#26323870")
    for node in ("CIE XYZ", "RGB", "Spectral Distribution"):
        agraph.get_node(node.lower()).attr.update(
            shape="doublecircle",
            color="#673AB7FF",
            fillcolor="#673AB770",
            fontsize=30,
        )
    for node in (
        "ATD95",
        "CAM16",
        "CIECAM02",
        "Hunt",
        "Kim 2009",
        "LLAB",
        "Nayatani95",
        "RLAB",
        "ZCAM",
    ):
        agraph.get_node(node.lower()).attr.update(
            color="#00BCD4FF", fillcolor="#00BCD470"
        )

    agraph.draw(filename, prog=prog, args=args)

    return agraph



"""
Colour Blindness Plotting
=========================

Defines the colour blindness plotting objects:

-   :func:`colour.plotting.plot_cvd_simulation_Machado2009`
"""

from __future__ import annotations

import matplotlib.pyplot as plt

from colour.algebra import vector_dot
from colour.blindness import matrix_cvd_Machado2009
from colour.hints import (
    Any,
    ArrayLike,
    Dict,
    Floating,
    Literal,
    Optional,
    Tuple,
    Union,
    cast,
)
from colour.plotting import CONSTANTS_COLOUR_STYLE, plot_image, override_style
from colour.utilities import optional

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "plot_cvd_simulation_Machado2009",
]


@override_style()
def plot_cvd_simulation_Machado2009(
    RGB: ArrayLike,
    deficiency: Union[
        Literal["Deuteranomaly", "Protanomaly", "Tritanomaly"], str
    ] = "Protanomaly",
    severity: Floating = 0.5,
    M_a: Optional[ArrayLike] = None,
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Perform colour vision deficiency simulation on given *RGB* colourspace
    array using *Machado et al. (2009)* model.

    Parameters
    ----------
    RGB
        *RGB* colourspace array.
    deficiency
        Colour blindness / vision deficiency type.
    severity
        Severity of the colour vision deficiency in domain [0, 1].
    M_a
        Anomalous trichromacy matrix to use instead of Machado (2010)
        pre-computed matrix.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.artist`, :func:`colour.plotting.plot_image`,
        :func:`colour.plotting.render`},
        See the documentation of the previously listed definitions.

    Notes
    -----
    -  Input *RGB* array is expected to be linearly encoded.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.

    Examples
    --------
    >>> import numpy as np
    >>> RGB = np.random.rand(32, 32, 3)
    >>> plot_cvd_simulation_Machado2009(RGB)  # doctest: +ELLIPSIS
    (<Figure size ... with 1 Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_Plot_CVD_Simulation_Machado2009.png
        :align: center
        :alt: plot_cvd_simulation_Machado2009
    """

    M_a = cast(
        ArrayLike, optional(M_a, matrix_cvd_Machado2009(deficiency, severity))
    )

    settings: Dict[str, Any] = {
        "text_kwargs": {
            "text": f"Deficiency: {deficiency} - Severity: {severity}"
        }
    }
    settings.update(kwargs)

    return plot_image(
        CONSTANTS_COLOUR_STYLE.colour.colourspace.cctf_encoding(
            vector_dot(M_a, RGB)
        ),
        **settings,
    )



"""
CMCCAT2000 Chromatic Adaptation Model
=====================================

Defines the *CMCCAT2000* chromatic adaptation model objects:

-   :class:`colour.adaptation.InductionFactors_CMCCAT2000`
-   :class:`colour.VIEWING_CONDITIONS_CMCCAT2000`
-   :func:`colour.adaptation.chromatic_adaptation_forward_CMCCAT2000`
-   :func:`colour.adaptation.chromatic_adaptation_inverse_CMCCAT2000`
-   :func:`colour.adaptation.chromatic_adaptation_CMCCAT2000`

References
----------
-   :cite:`Li2002a` : Li, C., Luo, M. R., Rigg, B., & Hunt, R. W. G. (2002).
    CMC 2000 chromatic adaptation transform: CMCCAT2000. Color Research &
    Application, 27(1), 49-58. doi:10.1002/col.10005
-   :cite:`Westland2012k` : Westland, S., Ripamonti, C., & Cheung, V. (2012).
    CMCCAT2000. In Computational Colour Science Using MATLAB (2nd ed., pp.
    83-86). ISBN:978-0-470-66569-5
"""

from __future__ import annotations

import numpy as np
from typing import NamedTuple

from colour.adaptation import CAT_CMCCAT2000
from colour.algebra import vector_dot
from colour.hints import (
    ArrayLike,
    Floating,
    FloatingOrArrayLike,
    Literal,
    NDArray,
    Union,
)
from colour.utilities import (
    CaseInsensitiveMapping,
    as_float_array,
    from_range_100,
    to_domain_100,
    validate_method,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "CAT_INVERSE_CMCCAT2000",
    "InductionFactors_CMCCAT2000",
    "VIEWING_CONDITIONS_CMCCAT2000",
    "chromatic_adaptation_forward_CMCCAT2000",
    "chromatic_adaptation_inverse_CMCCAT2000",
    "chromatic_adaptation_CMCCAT2000",
]

CAT_INVERSE_CMCCAT2000: NDArray = np.linalg.inv(CAT_CMCCAT2000)
"""
Inverse *CMCCAT2000* chromatic adaptation transform.

CAT_INVERSE_CMCCAT2000
"""


class InductionFactors_CMCCAT2000(NamedTuple):
    """
    *CMCCAT2000* chromatic adaptation model induction factors.

    Parameters
    ----------
    F
        :math:`F` surround condition.

    References
    ----------
    :cite:`Li2002a`, :cite:`Westland2012k`
    """

    F: Floating


VIEWING_CONDITIONS_CMCCAT2000: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "Average": InductionFactors_CMCCAT2000(1),
        "Dim": InductionFactors_CMCCAT2000(0.8),
        "Dark": InductionFactors_CMCCAT2000(0.8),
    }
)
VIEWING_CONDITIONS_CMCCAT2000.__doc__ = """
Reference *CMCCAT2000* chromatic adaptation model viewing conditions.

References
----------
:cite:`Li2002a`, :cite:`Westland2012k`
"""


def chromatic_adaptation_forward_CMCCAT2000(
    XYZ: ArrayLike,
    XYZ_w: ArrayLike,
    XYZ_wr: ArrayLike,
    L_A1: FloatingOrArrayLike,
    L_A2: FloatingOrArrayLike,
    surround: InductionFactors_CMCCAT2000 = VIEWING_CONDITIONS_CMCCAT2000[
        "Average"
    ],
) -> NDArray:
    """
    Adapt given stimulus *CIE XYZ* tristimulus values from test viewing
    conditions to reference viewing conditions using *CMCCAT2000* forward
    chromatic adaptation model.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values of the stimulus to adapt.
    XYZ_w
        Test viewing condition *CIE XYZ* tristimulus values of the whitepoint.
    XYZ_wr
        Reference viewing condition *CIE XYZ* tristimulus values of the
        whitepoint.
    L_A1
        Luminance of test adapting field :math:`L_{A1}` in :math:`cd/m^2`.
    L_A2
        Luminance of reference adapting field :math:`L_{A2}` in :math:`cd/m^2`.
    surround
        Surround viewing conditions induction factors.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ_c* tristimulus values of the stimulus corresponding colour.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+
    | ``XYZ_w``  | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+
    | ``XYZ_wr`` | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ_c``  | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Li2002a`, :cite:`Westland2012k`

    Examples
    --------
    >>> XYZ = np.array([22.48, 22.74, 8.54])
    >>> XYZ_w = np.array([111.15, 100.00, 35.20])
    >>> XYZ_wr = np.array([94.81, 100.00, 107.30])
    >>> L_A1 = 200
    >>> L_A2 = 200
    >>> chromatic_adaptation_forward_CMCCAT2000(XYZ, XYZ_w, XYZ_wr, L_A1, L_A2)
    ... # doctest: +ELLIPSIS
    array([ 19.5269832...,  23.0683396...,  24.9717522...])
    """

    XYZ = to_domain_100(XYZ)
    XYZ_w = to_domain_100(XYZ_w)
    XYZ_wr = to_domain_100(XYZ_wr)
    L_A1 = as_float_array(L_A1)
    L_A2 = as_float_array(L_A2)

    RGB = vector_dot(CAT_CMCCAT2000, XYZ)
    RGB_w = vector_dot(CAT_CMCCAT2000, XYZ_w)
    RGB_wr = vector_dot(CAT_CMCCAT2000, XYZ_wr)

    D = surround.F * (
        0.08 * np.log10(0.5 * (L_A1 + L_A2))
        + 0.76
        - 0.45 * (L_A1 - L_A2) / (L_A1 + L_A2)
    )

    D = np.clip(D, 0, 1)
    a = D * XYZ_w[..., 1] / XYZ_wr[..., 1]

    RGB_c = RGB * (
        a[..., np.newaxis] * (RGB_wr / RGB_w) + 1 - D[..., np.newaxis]
    )
    XYZ_c = vector_dot(CAT_INVERSE_CMCCAT2000, RGB_c)

    return from_range_100(XYZ_c)


def chromatic_adaptation_inverse_CMCCAT2000(
    XYZ_c: ArrayLike,
    XYZ_w: ArrayLike,
    XYZ_wr: ArrayLike,
    L_A1: FloatingOrArrayLike,
    L_A2: FloatingOrArrayLike,
    surround: InductionFactors_CMCCAT2000 = VIEWING_CONDITIONS_CMCCAT2000[
        "Average"
    ],
) -> NDArray:
    """
    Adapt given stimulus corresponding colour *CIE XYZ* tristimulus values
    from reference viewing conditions to test viewing conditions using
    *CMCCAT2000* inverse chromatic adaptation model.

    Parameters
    ----------
    XYZ_c
        *CIE XYZ* tristimulus values of the stimulus to adapt.
    XYZ_w
        Test viewing condition *CIE XYZ* tristimulus values of the whitepoint.
    XYZ_wr
        Reference viewing condition *CIE XYZ* tristimulus values of the
        whitepoint.
    L_A1
        Luminance of test adapting field :math:`L_{A1}` in :math:`cd/m^2`.
    L_A2
        Luminance of reference adapting field :math:`L_{A2}` in :math:`cd/m^2`.
    surround
        Surround viewing conditions induction factors.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ_c* tristimulus values of the adapted stimulus.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ_c``  | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+
    | ``XYZ_w``  | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+
    | ``XYZ_wr`` | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Li2002a`, :cite:`Westland2012k`

    Examples
    --------
    >>> XYZ_c = np.array([19.53, 23.07, 24.97])
    >>> XYZ_w = np.array([111.15, 100.00, 35.20])
    >>> XYZ_wr = np.array([94.81, 100.00, 107.30])
    >>> L_A1 = 200
    >>> L_A2 = 200
    >>> chromatic_adaptation_inverse_CMCCAT2000(XYZ_c, XYZ_w, XYZ_wr, L_A1,
    ...                                         L_A2)
    ... # doctest: +ELLIPSIS
    array([ 22.4839876...,  22.7419485...,   8.5393392...])
    """

    XYZ_c = to_domain_100(XYZ_c)
    XYZ_w = to_domain_100(XYZ_w)
    XYZ_wr = to_domain_100(XYZ_wr)
    L_A1 = as_float_array(L_A1)
    L_A2 = as_float_array(L_A2)

    RGB_c = vector_dot(CAT_CMCCAT2000, XYZ_c)
    RGB_w = vector_dot(CAT_CMCCAT2000, XYZ_w)
    RGB_wr = vector_dot(CAT_CMCCAT2000, XYZ_wr)

    D = surround.F * (
        0.08 * np.log10(0.5 * (L_A1 + L_A2))
        + 0.76
        - 0.45 * (L_A1 - L_A2) / (L_A1 + L_A2)
    )

    D = np.clip(D, 0, 1)
    a = D * XYZ_w[..., 1] / XYZ_wr[..., 1]

    RGB = RGB_c / (
        a[..., np.newaxis] * (RGB_wr / RGB_w) + 1 - D[..., np.newaxis]
    )
    XYZ = vector_dot(CAT_INVERSE_CMCCAT2000, RGB)

    return from_range_100(XYZ)


def chromatic_adaptation_CMCCAT2000(
    XYZ: ArrayLike,
    XYZ_w: ArrayLike,
    XYZ_wr: ArrayLike,
    L_A1: FloatingOrArrayLike,
    L_A2: FloatingOrArrayLike,
    surround: InductionFactors_CMCCAT2000 = VIEWING_CONDITIONS_CMCCAT2000[
        "Average"
    ],
    direction: Union[Literal["Forward", "Inverse"], str] = "Forward",
) -> NDArray:
    """
    Adapt given stimulus *CIE XYZ* tristimulus values using given viewing
    conditions.

    This definition is a convenient wrapper around
    :func:`colour.adaptation.chromatic_adaptation_forward_CMCCAT2000` and
    :func:`colour.adaptation.chromatic_adaptation_inverse_CMCCAT2000`.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values of the stimulus to adapt.
    XYZ_w
        Source viewing condition *CIE XYZ* tristimulus values of the
        whitepoint.
    XYZ_wr
        Target viewing condition *CIE XYZ* tristimulus values of the
        whitepoint.
    L_A1
        Luminance of test adapting field :math:`L_{A1}` in :math:`cd/m^2`.
    L_A2
        Luminance of reference adapting field :math:`L_{A2}` in :math:`cd/m^2`.
    surround
        Surround viewing conditions induction factors.
    direction
        Chromatic adaptation direction.

    Returns
    -------
    :class:`numpy.ndarray`
        Adapted stimulus *CIE XYZ* tristimulus values.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+
    | ``XYZ_w``  | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+
    | ``XYZ_wr`` | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Li2002a`, :cite:`Westland2012k`

    Examples
    --------
    >>> XYZ = np.array([22.48, 22.74, 8.54])
    >>> XYZ_w = np.array([111.15, 100.00, 35.20])
    >>> XYZ_wr = np.array([94.81, 100.00, 107.30])
    >>> L_A1 = 200
    >>> L_A2 = 200
    >>> chromatic_adaptation_CMCCAT2000(
    ...     XYZ, XYZ_w, XYZ_wr, L_A1, L_A2, direction='Forward')
    ... # doctest: +ELLIPSIS
    array([ 19.5269832...,  23.0683396...,  24.9717522...])

    Using the *CMCCAT2000* inverse model:

    >>> XYZ = np.array([19.52698326, 23.06833960, 24.97175229])
    >>> XYZ_w = np.array([111.15, 100.00, 35.20])
    >>> XYZ_wr = np.array([94.81, 100.00, 107.30])
    >>> L_A1 = 200
    >>> L_A2 = 200
    >>> chromatic_adaptation_CMCCAT2000(
    ...     XYZ, XYZ_w, XYZ_wr, L_A1, L_A2, direction='Inverse')
    ... # doctest: +ELLIPSIS
    array([ 22.48,  22.74,   8.54])
    """

    direction = validate_method(
        direction,
        ["Forward", "Inverse"],
        '"{0}" direction is invalid, it must be one of {1}!',
    )

    if direction == "forward":
        return chromatic_adaptation_forward_CMCCAT2000(
            XYZ, XYZ_w, XYZ_wr, L_A1, L_A2, surround
        )
    else:
        return chromatic_adaptation_inverse_CMCCAT2000(
            XYZ, XYZ_w, XYZ_wr, L_A1, L_A2, surround
        )



"""
Defines the unit tests for the
:mod:`colour.models.rgb.transfer_functions.gamma` module.
"""

import numpy as np
import unittest

from colour.models.rgb.transfer_functions import gamma_function
from colour.utilities import ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestGammaFunction",
]


class TestGammaFunction(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.gamma.gamma_function`
    definition unit tests methods.
    """

    def test_gamma_function(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.gamma.\
gamma_function` definition.
        """

        self.assertAlmostEqual(gamma_function(0.0, 2.2), 0.0, places=7)

        self.assertAlmostEqual(
            gamma_function(0.18, 2.2), 0.022993204992707, places=7
        )

        self.assertAlmostEqual(
            gamma_function(0.022993204992707, 1.0 / 2.2), 0.18, places=7
        )

        self.assertAlmostEqual(
            gamma_function(-0.18, 2.0), 0.0323999999999998, places=7
        )

        np.testing.assert_array_equal(gamma_function(-0.18, 2.2), np.nan)

        self.assertAlmostEqual(
            gamma_function(-0.18, 2.2, "Mirror"), -0.022993204992707, places=7
        )

        self.assertAlmostEqual(
            gamma_function(-0.18, 2.2, "Preserve"), -0.18, places=7
        )

        self.assertAlmostEqual(
            gamma_function(-0.18, 2.2, "Clamp"), 0, places=7
        )

        np.testing.assert_array_equal(gamma_function(-0.18, -2.2), np.nan)

        self.assertAlmostEqual(
            gamma_function(0.0, -2.2, "Mirror"), 0.0, places=7
        )

        self.assertAlmostEqual(
            gamma_function(0.0, 2.2, "Preserve"), 0.0, places=7
        )

        self.assertAlmostEqual(gamma_function(0.0, 2.2, "Clamp"), 0, places=7)

    def test_n_dimensional_gamma_function(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.gamma.\
gamma_function` definition n-dimensional arrays support.
        """

        a = 0.18
        a_p = gamma_function(a, 2.2)

        a = np.tile(a, 6)
        a_p = np.tile(a_p, 6)
        np.testing.assert_almost_equal(gamma_function(a, 2.2), a_p, decimal=7)

        a = np.reshape(a, (2, 3))
        a_p = np.reshape(a_p, (2, 3))
        np.testing.assert_almost_equal(gamma_function(a, 2.2), a_p, decimal=7)

        a = np.reshape(a, (2, 3, 1))
        a_p = np.reshape(a_p, (2, 3, 1))
        np.testing.assert_almost_equal(gamma_function(a, 2.2), a_p, decimal=7)

        a = -0.18
        a_p = -0.022993204992707
        np.testing.assert_almost_equal(
            gamma_function(a, 2.2, "Mirror"), a_p, decimal=7
        )

        a = np.tile(a, 6)
        a_p = np.tile(a_p, 6)
        np.testing.assert_almost_equal(
            gamma_function(a, 2.2, "Mirror"), a_p, decimal=7
        )

        a = np.reshape(a, (2, 3))
        a_p = np.reshape(a_p, (2, 3))
        np.testing.assert_almost_equal(
            gamma_function(a, 2.2, "Mirror"), a_p, decimal=7
        )

        a = np.reshape(a, (2, 3, 1))
        a_p = np.reshape(a_p, (2, 3, 1))
        np.testing.assert_almost_equal(
            gamma_function(a, 2.2, "Mirror"), a_p, decimal=7
        )

        a = -0.18
        a_p = -0.18
        np.testing.assert_almost_equal(
            gamma_function(a, 2.2, "Preserve"), a_p, decimal=7
        )

        a = np.tile(a, 6)
        a_p = np.tile(a_p, 6)
        np.testing.assert_almost_equal(
            gamma_function(a, 2.2, "Preserve"), a_p, decimal=7
        )

        a = np.reshape(a, (2, 3))
        a_p = np.reshape(a_p, (2, 3))
        np.testing.assert_almost_equal(
            gamma_function(a, 2.2, "Preserve"), a_p, decimal=7
        )

        a = np.reshape(a, (2, 3, 1))
        a_p = np.reshape(a_p, (2, 3, 1))
        np.testing.assert_almost_equal(
            gamma_function(a, 2.2, "Preserve"), a_p, decimal=7
        )

        a = -0.18
        a_p = 0.0
        np.testing.assert_almost_equal(
            gamma_function(a, 2.2, "Clamp"), a_p, decimal=7
        )

        a = np.tile(a, 6)
        a_p = np.tile(a_p, 6)
        np.testing.assert_almost_equal(
            gamma_function(a, 2.2, "Clamp"), a_p, decimal=7
        )

        a = np.reshape(a, (2, 3))
        a_p = np.reshape(a_p, (2, 3))
        np.testing.assert_almost_equal(
            gamma_function(a, 2.2, "Clamp"), a_p, decimal=7
        )

        a = np.reshape(a, (2, 3, 1))
        a_p = np.reshape(a_p, (2, 3, 1))
        np.testing.assert_almost_equal(
            gamma_function(a, 2.2, "Clamp"), a_p, decimal=7
        )

    @ignore_numpy_errors
    def test_nan_gamma_function(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.gamma.\
gamma_function` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]

        for case in cases:
            gamma_function(case, case)


if __name__ == "__main__":
    unittest.main()



"""Showcases *ANSI/IES TM-30-18 Colour Rendition Report* plotting examples."""

import colour
from colour.plotting import (
    plot_single_sd_colour_rendition_report,
    colour_style,
)
from colour.utilities import message_box

message_box("ANSI/IES TM-30-18 Colour Rendition Report")

colour_style()

sd = colour.SDS_ILLUMINANTS["FL2"]

message_box('Plotting a full "ANSI/IES TM-30-18 Colour Rendition Report".')
plot_single_sd_colour_rendition_report(sd)

print("\n")

message_box(
    'Plotting an intermediate "ANSI/IES TM-30-18 Colour Rendition Report".'
)
plot_single_sd_colour_rendition_report(sd, "Intermediate")

print("\n")

message_box('Plotting a simple "ANSI/IES TM-30-18 Colour Rendition Report".')
plot_single_sd_colour_rendition_report(sd, "Simple")



"""
ITU-R BT.470 Colourspaces
=========================

Defines the *ITU-R BT.470* colourspaces:

-   :attr:`colour.models.RGB_COLOURSPACE_BT470_525`.
-   :attr:`colour.models.RGB_COLOURSPACE_BT470_625`.

References
----------
-   :cite:`InternationalTelecommunicationUnion1998a` : International
    Telecommunication Union. (1998). Recommendation ITU-R BT.470-6 -
    CONVENTIONAL TELEVISION SYSTEMS (pp. 1-36).
    http://www.itu.int/dms_pubrec/itu-r/rec/bt/\
R-REC-BT.470-6-199811-S!!PDF-E.pdf
"""

from __future__ import annotations

import numpy as np
from functools import partial

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import NDArray
from colour.models.rgb import (
    RGB_Colourspace,
    gamma_function,
    normalised_primary_matrix,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "PRIMARIES_BT470_525",
    "CCS_WHITEPOINT_BT470_525",
    "WHITEPOINT_NAME_BT470_525",
    "MATRIX_BT470_525_TO_XYZ",
    "MATRIX_XYZ_TO_BT470_525",
    "RGB_COLOURSPACE_BT470_525",
    "PRIMARIES_BT470_625",
    "CCS_WHITEPOINT_BT470_625",
    "WHITEPOINT_NAME_BT470_625",
    "MATRIX_BT470_625_TO_XYZ",
    "MATRIX_XYZ_TO_BT470_625",
    "RGB_COLOURSPACE_BT470_625",
]

PRIMARIES_BT470_525: NDArray = np.array(
    [
        [0.6700, 0.3300],
        [0.2100, 0.7100],
        [0.1400, 0.0800],
    ]
)
"""*ITU-R BT.470 - 525* colourspace primaries."""

WHITEPOINT_NAME_BT470_525: str = "C"
"""*ITU-R BT.470 - 525* colourspace whitepoint name."""

CCS_WHITEPOINT_BT470_525: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
][WHITEPOINT_NAME_BT470_525]
"""*ITU-R BT.470 - 525* colourspace whitepoint chromaticity coordinates."""

MATRIX_BT470_525_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_BT470_525, CCS_WHITEPOINT_BT470_525
)
"""*ITU-R BT.470 - 525* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_BT470_525: NDArray = np.linalg.inv(MATRIX_BT470_525_TO_XYZ)
"""*CIE XYZ* tristimulus values to *ITU-R BT.470 - 525* colourspace matrix."""

RGB_COLOURSPACE_BT470_525: RGB_Colourspace = RGB_Colourspace(
    "ITU-R BT.470 - 525",
    PRIMARIES_BT470_525,
    CCS_WHITEPOINT_BT470_525,
    WHITEPOINT_NAME_BT470_525,
    MATRIX_BT470_525_TO_XYZ,
    MATRIX_XYZ_TO_BT470_525,
    partial(gamma_function, exponent=1 / 2.8),
    partial(gamma_function, exponent=2.8),
)
RGB_COLOURSPACE_BT470_525.__doc__ = """
*ITU-R BT.470 - 525* colourspace.

References
----------
:cite:`InternationalTelecommunicationUnion1998a`
"""

PRIMARIES_BT470_625: NDArray = np.array(
    [
        [0.6400, 0.3300],
        [0.2900, 0.6000],
        [0.1500, 0.0600],
    ]
)
"""*ITU-R BT.470 - 625* colourspace primaries."""

WHITEPOINT_NAME_BT470_625: str = "D65"
"""*ITU-R BT.470 - 625* colourspace whitepoint name."""

CCS_WHITEPOINT_BT470_625: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
][WHITEPOINT_NAME_BT470_625]
"""*ITU-R BT.470 - 625* colourspace whitepoint chromaticity coordinates."""

MATRIX_BT470_625_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_BT470_625, CCS_WHITEPOINT_BT470_625
)
"""*ITU-R BT.470 - 625* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_BT470_625: NDArray = np.linalg.inv(MATRIX_BT470_625_TO_XYZ)
"""*CIE XYZ* tristimulus values to *ITU-R BT.470 - 625* colourspace matrix."""

RGB_COLOURSPACE_BT470_625: RGB_Colourspace = RGB_Colourspace(
    "ITU-R BT.470 - 625",
    PRIMARIES_BT470_625,
    CCS_WHITEPOINT_BT470_625,
    WHITEPOINT_NAME_BT470_625,
    MATRIX_BT470_625_TO_XYZ,
    MATRIX_XYZ_TO_BT470_625,
    partial(gamma_function, exponent=1 / 2.8),
    partial(gamma_function, exponent=2.8),
)
RGB_COLOURSPACE_BT470_625.__doc__ = """
*ITU-R BT.470 - 625* colourspace.

References
----------
:cite:`InternationalTelecommunicationUnion1998a`
"""
